//
// Created by nicolas on 20/03/18.
//

#include <stddef.h>
#include <stdlib.h>
#include <assert.h>
#include "debug.h"
#include "queue.h"
#include "mutex.h"

struct st_node {
    void* item;
    struct st_node* next;
};
struct st_queue {
    mutex_t* lock;
    size_t size;
    struct st_node* start;
    struct st_node* end;
};

queue_t* queue_create() {
    queue_t* q = malloc(sizeof(struct st_queue));
    assert(q != 0);
    q->lock = mutex_create();
    q->size = 0;
    q->start = 0;
    q->end = 0;
    return q;
}

void queue_enqueue(queue_t* q, void* item) {
    struct st_node* node = malloc(sizeof(struct st_node));
    assert(node != 0);
    node->item = item;
    node->next = 0;

    mutex_lock(q->lock);
    if (q->end) {
        q->end->next = node;
    }
    if (q->start == 0) {
        q->start = node;
    }
    q->end = node;
    q->size++;
    mutex_unlock(q->lock);
}

void* queue_dequeue(queue_t* q) {
    mutex_lock(q->lock);
    if (q->start == 0) {
        mutex_unlock(q->lock);
        return 0;
    }
    struct st_node* node = q->start;
    void* item = node->item;
    q->start = node->next;
    if (q->end == node) {
        q->end = 0;
    }
    q->size--;
    mutex_unlock(q->lock);
    free(node);
    return item;
}

void* queue_peek(queue_t* q) {
    mutex_lock(q->lock);
    void* item = q->start ? q->start->item : 0;
    mutex_unlock(q->lock);
    return item;
}

inline size_t queue_size(queue_t* q) {
    mutex_lock(q->lock);
    size_t size = q->size;
    mutex_unlock(q->lock);
    return size;
}

void queue_destroy(queue_t* q) {
    mutex_lock(q->lock);
    struct st_node* p = q->start;
    while(p) {
        struct st_node* next = p->next;
        free(p);
        p = next;
    }
    mutex_unlock(q->lock);
    mutex_destroy(q->lock);
    free(q);
}

void* queue_remove(queue_t* q, find_predicate predicate, void* arg) {
    void* found = 0;
    mutex_lock(q->lock);
    struct st_node* p = q->start;
    struct st_node* prev = 0;
    while(p) {
        if (predicate(p->item, arg) == BOOL_TRUE) {
            found = p->item;
            if (prev) {
                prev->next = p->next;
            } else {
                q->start = p->next;
            }
            if (q->end == p) {
                q->end = prev;
            }
            free(p);
            break;
        }
        prev = p;
        p = p->next;
    }
    mutex_unlock(q->lock);
    return found;
}

void* queue_find(queue_t* q, find_predicate predicate, void* arg) {
    void* found = 0;
    mutex_lock(q->lock);
    struct st_node* p = q->start;
    while(p) {
        if (predicate(p->item, arg) == BOOL_TRUE) {
            found = p->item;
            break;
        }
        p = p->next;
    }
    mutex_unlock(q->lock);
    return found;
}
